/*
    SSTP Bottle Client for MacOS X, using protocol SLPP2.0
    Copyright (C) 2001-2002 phonohawk.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package gui.logbrowser;
import gui.logbrowser.*;
import log.*;
import core.*;
import com.apple.cocoa.foundation.*;
import com.apple.cocoa.application.*;
import java.io.*;
import java.net.*;
import java.util.*;
import java.util.zip.*;

public class LogBrowserController extends NSWindowController
{
    static LogBrowserController _sharedInstance = null;
    
    static String user = NSPathUtilities.stringByExpandingTildeInPath("~");
    
    //static String[] long_month_names = {"January","February","March","April","May","June","July","August","September","October","November","December"};
    static String[] short_month_names = {"Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"};
    
    static String valid_uri_special_chars = ".+%&#~?_/-=";
    
	NSBox box_top;
    NSBrowser browse_days;
    NSButton btn_replay;
    NSButton btn_vote;
    NSButton btn_agree;
    NSButton btn_copyurl;
    NSButton btn_jumptourl;
    NSTextField fld_day;
	NSTextField fld_group;
	NSTextField fld_quantity;
    NSTableView table_scripts;
	NSTableView table_groups;
	NSDrawer drawer_groups;
    
    LogBrowserListViewDataSource listDataSource;
	GroupListDataSource groupDataSource;
    
    public static LogBrowserController sharedLogBrowser()
    {
        if (_sharedInstance == null)
        {
            _sharedInstance = new LogBrowserController();
        }
        
        return _sharedInstance;
    }
    
    private LogBrowserController()
    {
        super("LogWindow");
        loadWindow();
        
		window().setFrameUsingName("logwindow_frame");
        window().setFrameAutosaveName("logwindow_frame");
        table_scripts.setAutosaveName("logwindow_table_revision_2");
		table_scripts.setAutosaveTableColumns(true);
		table_groups.setAutosaveName("logwindow.table_groups.revision.0");
		table_groups.setAutosaveTableColumns(true);
        browse_days.setTakesTitleFromPreviousColumn(false);
		
		NSUserDefaults defaults = NSUserDefaults.standardUserDefaults();
		if (defaults.integerForKey("logwindow.drawer_groups.open") == 1) {
			drawer_groups.open();
		}
		else {
			drawer_groups.close();
		}
		
		int box_top_height = defaults.integerForKey("logwindow.box_top.height");
		if (box_top_height > 0) {
			box_top.setFrameSize(new NSSize(box_top.frame().width(),box_top_height));
		}
        
        listDataSource = new LogBrowserListViewDataSource(this);
        table_scripts.setDataSource(listDataSource);
		
		groupDataSource = new GroupListDataSource();
		table_groups.setDataSource(groupDataSource);
        
        //browse_days.setAcceptsArrowKeys(true);
        browserSelectionHasChanged(null);
        
        buttons_validate();
    }
    
    public String browserPath()
    {
        return browse_days.path();
    }
	
	public void updateIndicators()
	{
		Object[] group = getSelectedGroup();
		if (group == null) {
			fld_group.setStringValue("");
			fld_quantity.setStringValue("");
		}
		else {
			fld_group.setStringValue((String)group[0]);
			fld_quantity.setIntValue(((Vector)group[1]).size());
		}
	}
    
    public void reloadTable()
    {
		table_groups.reloadData();
        table_scripts.reloadData();
		
        table_scripts.scrollRowToVisible(table_scripts.numberOfRows() - 1);
		//table_scripts.display();
		
		updateIndicators();
    }
    
    public void reloadTableIfShowing(int year,int month,int day)
    {
		// ̃Xgɕ\Ă̂w肳ꂽtƓ烊[hB
		String path = browse_days.path();
		try {
			int current_year = Integer.parseInt(path.substring(1,5));
			int current_month = Integer.parseInt(path.substring(6,8));
			int current_day = Integer.parseInt(path.substring(9,11));
			if (year == current_year && month == current_month && day == current_day) {
				// change toveta 2002.09.16
                // deselectAllĂȂƃ[hɃNAہAƂǂO
                NSEnumerator enu;
                Vector rows = new Vector();
                Number row;
                // Enumerator̒gtable𑀍삵Ă܂ƕۏ؂Ȃ?̂Vectorɑޔ
                enu = table_scripts.selectedRowEnumerator();
                row = (Number)enu.nextElement();
                while(row != null){
                    rows.add(row);
                    row = (Number)enu.nextElement();
                }
                table_scripts.deselectAll(this);
				browserSelectionHasChanged(null);
                for(int i=0,size=rows.size(); i < size; i++){
                    row = (Number)rows.elementAt(i);
                    table_scripts.selectRow(row.intValue(),true);
                }
			}
		} catch (Exception e) {}
		
		updateIndicators();
    }
	
	public Object[] getSelectedGroup()
	{
		// ݑIĂO[vƁA̒gԂB
		// IĂȂnullԂB
		// result: <Object[]> , [0]:(String)name [1]:(Vector)bottles
		
		LogData holded = LogDataServer.sharedLogServer().getHolded();
		if (holded == null) {
			return null;
		}
		
		Vector groups = holded.getGroups(); // <Object[]> , [0]:(String)name [1]:(Vector)bottles
		
		if (table_groups.numberOfSelectedRows() == 1) {
			return (Object[])groups.elementAt(table_groups.selectedRow());
		}
		else {
			return null;
		}
	}
    
    // Actions
    public void browserSelectionHasChanged(Object sender)
    {
        // IĂ̂dayicolumn 2jȂ牺̃Xgɒg\B
        if (browse_days.selectedColumn() == 2)
        {
            String path = browse_days.path();
            String year = path.substring(1,5);
            String month = path.substring(6,8);
            String day = path.substring(9,11);
            String month_str;
            try{ month_str = short_month_names[ Integer.parseInt(month) - 1 ]; }catch(Exception e){ month_str = month; }
            String mdy = month_str+'/'+day+'/'+year;
            fld_day.setStringValue(mdy);
            
			try {
				LogDataServer.sharedLogServer().hold(
										Integer.parseInt(year),
										Integer.parseInt(month),
										Integer.parseInt(day));
			}
			catch (NumberFormatException e) {}
        }
        else
        {
            fld_day.setStringValue("");
            
			LogDataServer.sharedLogServer().unhold();
        }
		
		table_groups.reloadData();
		table_scripts.reloadData();
		
		updateIndicators();
    }
	
	public void groupSelectionHasChanged(Object sender)
	{
		table_scripts.reloadData();
        table_scripts.scrollRowToVisible(table_scripts.numberOfRows() - 1);
		
		updateIndicators();
	}
    
    public void replay(Object sender)
    {
        LogData log = LogDataServer.sharedLogServer().getHolded();
		if (log == null) return;
        
        int n_rows = table_scripts.numberOfRows();
        for (int i = 0;i < n_rows;i++)
        {
            if (!table_scripts.isRowSelected(i)) continue;
            
            Hashtable info = log.elementAt(i);
            replay_bottle(info);
        }
    }
	
	public static void replay_bottle(Hashtable bottle)
	{
		String app_name = NSBundle.mainBundle().localizedStringForKey("CFBundleShortVersionString","BotCli","InfoPlist");
		String date = (String)bottle.get("date");
		String time = (String)bottle.get("time");
		String ch = (String)bottle.get("channel"); if (ch == null) ch = "";
		String script = (String)bottle.get("script"); if (script == null) script = "";
		String ghost = (String)bottle.get("ghost");
		String ifghost = (String)bottle.get("ifghost");
		
		// BotCli:replay:MM/DD:[HH:MM]:<channel>:<ifghost>
		String sender = "BotCli:replay:"+date.substring(0,2)+"/"+date.substring(2,4)+":["+time.substring(0,2)+":"+time.substring(2,4)+"]:"+ch+(ifghost==null?"":":"+ifghost);
		
		Vector req = new Vector();
		req.add("SEND SSTP/1.4");
		
		req.add("Sender: "+sender);
		if (ifghost != null) req.add("IfGhost: "+ifghost);
		else if (ghost != null) req.add("IfGhost: "+ghost);
		req.add("Script: "+ScriptAnalyzer.expandYenURL(script));
	
		Foundation.sharedFoundation().getQueue().push(req);
	}
    
    public void vote(Object sender)
    {
		LogData log = LogDataServer.sharedLogServer().getHolded();
        if (log == null) return;
	
		if (Foundation.sharedFoundation().slpp_connected())
		{
			if (table_scripts.numberOfSelectedRows() == 1)
			{
				Hashtable info = log.elementAt(table_scripts.selectedRow());
				if (info.get("mid") != null)
				{
					Foundation.sharedFoundation().getSlppConnection().vote((String)info.get("mid"),"Vote");
				}
			}
		}
    }
    
    public void agree(Object sender)
    {
		LogData log = LogDataServer.sharedLogServer().getHolded();
        if (log == null) return;
	
		if (Foundation.sharedFoundation().slpp_connected())
		{
			if (table_scripts.numberOfSelectedRows() == 1)
			{
				Hashtable info = log.elementAt(table_scripts.selectedRow());
				if (info.get("mid") != null)
				{
					Foundation.sharedFoundation().getSlppConnection().vote((String)info.get("mid"),"Agree");
				}
			}
		}
    }
    
    public void copy_url(Object sender)
    {
		// WFly[Xg{[hɃZbg
		String url_string = getURLFromSelectedBottle();
		if (url_string == null) return;
		
		NSPasteboard gen_p = NSPasteboard.generalPasteboard();
		NSMutableArray types = new NSMutableArray();
		types.addObject(NSPasteboard.StringPboardType);
		gen_p.declareTypes(types,null);
		gen_p.setStringForType(url_string,NSPasteboard.StringPboardType);
    }
    
    public void jump_to_url(Object sender)
    {
		// OSoRŃuEUɂURLJ
		String url_string = getURLFromSelectedBottle();
		if (url_string == null) return;
		
		try
		{
			NSWorkspace.sharedWorkspace().openURL(new URL(url_string));
		}
		catch (Exception e){}
    }
    
    public String getURLFromSelectedBottle()
    {
		// oȂnullԂB
		
		LogData log = LogDataServer.sharedLogServer().getHolded();
		if (log == null) return null;
		
		if (table_scripts.numberOfSelectedRows() == 1)
		{
			Hashtable info = log.elementAt(table_scripts.selectedRow());
			if (info.get("script") != null)
			{
				String script = (String)info.get("script");
				
				// http://TB
				int blockstart = script.indexOf("http://");
				if (blockstart == -1) return null;
				
				// At@xbgAAvalid_uri_special_charsȊO̕ɏoA܂͕񂪏I܂ł͈̔͂
				// URLubNƂB
				int blockend = blockstart + 7;
				int len = script.length();
				while (blockend < len)
				{
					char current = script.charAt(blockend);
					if ((current >= 'A' && current <= 'Z') || 
						(current >= 'a' && current <= 'z') ||
						Character.isDigit(current) ||
						valid_uri_special_chars.indexOf(current) != -1)
					{
						blockend++;
					}
					else
					{
						break;
					}
				}
				if (blockstart == blockend) return null; // http://̌ɑĂȂ
				String url_string = script.substring(blockstart,blockend);
				
				return url_string;
			}
		}
		
		return null;
    }
    
    // Delegete of NSBrowser
    public int browserNumberOfRowsInColumn(NSBrowser sender,int column)
    {
		String path = browse_days.path();
		
		if (column == 0) {
			return LogDataServer.logged_years().size();
		}
		else if (column == 1) {
            String year = path.substring(1,5);
			return LogDataServer.logged_months(year).size();
		}
		else if (column == 2) {
			String year = path.substring(1,5);
            String month = path.substring(6,8);
			return LogDataServer.logged_days(year,month).size();
		}
		else {
			return 0;
		}
    }
    
    public void browserWillDisplayCell(NSBrowser sender,Object cell,int row,int column)
    {
		String path = browse_days.path();
		
		Vector vec;
		if (column == 0) {
			vec = LogDataServer.logged_years();
		}
		else if (column == 1) {
            String year = path.substring(1,5);
			vec = LogDataServer.logged_months(year);
		}
		else if (column == 2) {
			String year = path.substring(1,5);
            String month = path.substring(6,8);
			vec = LogDataServer.logged_days(year,month);
		}
		else {
			return;
		}
		
        NSBrowserCell bcell = (NSBrowserCell)cell;
		bcell.setStringValue((String)vec.elementAt(row));
		bcell.setLeaf(column == 2);
    }
    
    public String browserTitleOfColumn(NSBrowser sender,int column)
    {
        switch (column)
        {
            case 0: return "Year";
            case 1: return "Month";
            case 2: return "Day";
            default: return "";
        }
    }
    
    public void buttons_validate()
	{
		btn_replay.setEnabled(table_scripts.numberOfSelectedRows() > 0);
		
		LogData log = LogDataServer.sharedLogServer().getHolded();
        if (log == null) return;
	
		// Vote{^Agree{^
		boolean votebutton_enable = false;
		if (Foundation.sharedFoundation().slpp_connected())
		{
			if (table_scripts.numberOfSelectedRows() == 1)
			{
				Object[] group = getSelectedGroup();
				Hashtable info = (Hashtable)((Vector)group[1]).elementAt(table_scripts.selectedRow());
				
				// ̃{gmid݂Γ[/Ӊ\B
				if (info.get("mid") != null) {
					votebutton_enable = true;
				}
			}
		}
		btn_vote.setEnabled(votebutton_enable);
		btn_agree.setEnabled(votebutton_enable);
	
		// Copy URL,Jump to URL{^
		boolean urlbutton_enable = false;
		if (table_scripts.numberOfSelectedRows() == 1)
		{
			Hashtable info = log.elementAt(table_scripts.selectedRow());
			if (info.get("script") != null)
			{
				String script = (String)info.get("script");
				
				urlbutton_enable = (script.indexOf("http://") != -1);
			}
		}
		btn_copyurl.setEnabled(urlbutton_enable);
		btn_jumptourl.setEnabled(urlbutton_enable);
    }
    
    // Delegete of NSTableView
    public void tableViewSelectionDidChange(NSNotification notification)
    {
        // notification͖
        buttons_validate();
    }
	
	// Delegate of NSDrawer
	public void drawerDidClose(NSNotification notification)
	{
		drawerStateHasChanged(notification);
	}
	public void drawerDidOpen(NSNotification notification)
	{
		drawerStateHasChanged(notification);
	}
	protected void drawerStateHasChanged(NSNotification notification)
	{
		if (notification.object() == drawer_groups) {
			int drawer_state = (drawer_groups.state() == NSDrawer.OpenState ? 1 : 0);
			
			NSUserDefaults defaults = NSUserDefaults.standardUserDefaults();
			defaults.setIntegerForKey(drawer_state,"logwindow.drawer_groups.open");
			defaults.synchronize();
		}
	}
	
	// Delegate of NSWindow
	public void windowWillClose(NSNotification aNotification)
	{
		if (aNotification.object() == window()) {
			int box_top_height = (int)box_top.frame().height();
			
			NSUserDefaults defaults = NSUserDefaults.standardUserDefaults();
			defaults.setIntegerForKey(box_top_height,"logwindow.box_top.height");
			defaults.synchronize();
		}
	}
}
