/*
	$Id: HttpSession.java,v 1.1.1.1 2003/04/08 14:55:47 admin Exp $
	-----
	
    SSTP Bottle Client for MacOS X, using protocol SLPP2.0
    Copyright (C) 2001-2003 phonohawk.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package log.logviewer;
import core.*;
import log.*;
import gui.logbrowser.*;
import log.logviewer.*;
import java.net.*;
import java.io.*;
import java.util.*;

public class HttpSession extends Thread
{
	public static final int RANGE_STEP = 100;
	
	Socket s;
	BufferedReader in;
	BufferedWriter out;
	
	public HttpSession(Socket s)
	{
		try {
			this.s = s;
			in = new BufferedReader(new InputStreamReader(s.getInputStream()));
			out = new BufferedWriter(new OutputStreamWriter(s.getOutputStream(),"UTF-8"));
		}
		catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	public void run()
	{
		try {
			StringTokenizer header_st = new StringTokenizer(in.readLine()," "); // GET / HTTP/1.0
			String method = header_st.nextToken();
			String path = header_st.nextToken();
			String http_version = header_st.nextToken();
			
			Hashtable headers = new Hashtable();
			while (true) {
				String header_line = in.readLine();
				if (header_line.equals("")) {
					break;
				}
				
				int colon_pos = header_line.indexOf(':');
				String key = header_line.substring(0,colon_pos).trim();
				String value = header_line.substring(colon_pos+1).trim();
				headers.put(key,value);
			}
			
			if (method.equals("GET")) {
				method_get(path,headers);
			}
			else {
				out.write("HTTP/1.0 405 Method Not Allowed\r\n\r\n");
				out.flush();
			}
		}
		catch (Exception e) {
			System.err.println(e.toString());
			e.printStackTrace();
		}
		finally {
			try {
				s.close();
			}
			catch (Exception e) {}
		}
	}
	
	public void method_get(String path,Hashtable headers) throws Exception
	{
		String query = null;
		int question_pos = path.indexOf('?');
		if (question_pos != -1) {
			query = path.substring(question_pos+1);
			path = path.substring(0,question_pos);
		}
		
		StringTokenizer path_st = new StringTokenizer(path,"/");
		String command;
		if (path_st.hasMoreTokens()) {
			command = path_st.nextToken();
		}
		else {
			command = "list";
		}
		
		
		if (command.equals("list")) {
			cmd_list(path_st);
		}
		else if (command.equals("css")) {
			cmd_css(path_st);
		}
		else if (command.equals("vote")) {
			cmd_vote(path_st);
		}
		else if (command.equals("agree")) {
			cmd_agree(path_st);
		}
		else if (command.equals("replay")) {
			cmd_replay(path_st,query);
		}
		else {
			err_404();
		}
	}
	
	public void cmd_list(StringTokenizer path_st) throws Exception
	{
		String selected_year = null;
		String selected_month = null;
		String selected_day = null;
		String range = null;
		int n_tokens = path_st.countTokens();
		if (n_tokens >= 1) {
			selected_year = path_st.nextToken();
		}
		if (n_tokens >= 2) {
			selected_month = path_st.nextToken();
		}
		if (n_tokens >= 3) {
			selected_day = path_st.nextToken();
		}
		if (n_tokens >= 4) {
			range = path_st.nextToken();
		}
		
		if (selected_year != null && selected_year.equals("today")) {
			range = selected_month;
			
			Calendar cal = Calendar.getInstance();
			int year = cal.get(Calendar.YEAR);
			int month = cal.get(Calendar.MONTH) + 1;
			int day = cal.get(Calendar.DAY_OF_MONTH);
			String expanded = "/list/"+year+"/"+(month < 10 ? "0" : "")+month+"/"+
				(day < 10 ? "0" : "")+day+"/"+(range == null ? "" : range);
			redirect_found(expanded);
		}
		else {
			if (n_tokens <= 2) {
				cmd_list_select(selected_year,selected_month);
			}
			else {
				cmd_list_bottles(selected_year,selected_month,selected_day,range);
			}
		}
	}
	
	public void cmd_list_select(String selected_year,String selected_month) throws Exception
	{
		TextTemplate tmpl_main = Httpd.tmpl_index.get("main");
		tmpl_main.set("title","Menu");
		
		TextTemplate tmpl_year_select = Httpd.tmpl_index.get("year-select");
		TextTemplate tmpl_year_link = Httpd.tmpl_index.get("year-link");
		TextTemplate tmpl_year_selected = Httpd.tmpl_index.get("year-selected");
		TextTemplate tmpl_month_select = Httpd.tmpl_index.get("month-select");
		TextTemplate tmpl_month_link = Httpd.tmpl_index.get("month-link");
		TextTemplate tmpl_month_selected = Httpd.tmpl_index.get("month-selected");
		TextTemplate tmpl_day_select = Httpd.tmpl_index.get("day-select");
		TextTemplate tmpl_day_link = Httpd.tmpl_index.get("day-link");
		StringBuffer years_html = new StringBuffer();
		for (Enumeration e = LogDataServer.logged_years().elements(); e.hasMoreElements(); ) {
			String year = (String)e.nextElement();
			
			if (year.equals(selected_year)) {
				tmpl_year_selected.set("year",year);
				years_html.append(tmpl_year_selected.evaluate()+"\n");
				
				tmpl_month_link.set("year",year);
				tmpl_day_link.set("year",year);
				StringBuffer months_html = new StringBuffer();
				for (Enumeration en = LogDataServer.logged_months(year).elements(); en.hasMoreElements(); ) {
					String month = (String)en.nextElement();
					
					if (month.equals(selected_month)) {
						tmpl_month_selected.set("month",month);
						months_html.append(tmpl_month_selected.evaluate()+"\n");
						
						tmpl_day_link.set("month",month);
						StringBuffer days_html = new StringBuffer();
						int day_count = 1;
						for (Enumeration enu = LogDataServer.logged_days(year,month).elements();
								enu.hasMoreElements();
								day_count++) {
							
							tmpl_day_link.set("day",(String)enu.nextElement());
							if (day_count % 10 == 0) {
								days_html.append(tmpl_day_link.evaluate()+"<br>\n");
							}
							else {
								days_html.append(tmpl_day_link.evaluate()+" ");
							}
						}
						
						tmpl_day_select.set("days",days_html.toString());
						months_html.append(tmpl_day_select.evaluate()+"\n");
					}
					else {
						tmpl_month_link.set("month",month);
						months_html.append(tmpl_month_link.evaluate()+"\n");
					}
				}
				
				tmpl_month_select.set("months",months_html.toString());
				years_html.append(tmpl_month_select.evaluate()+"\n");
			}
			else {
				tmpl_year_link.set("year",year);
				years_html.append(tmpl_year_link.evaluate()+"\n");
			}
		}
		tmpl_year_select.set("years",years_html.toString());
		
		tmpl_main.set("main-content",tmpl_year_select.evaluate());
		ok_200(tmpl_main.evaluate());
	}
	
	public void cmd_list_bottles(String selected_year,String selected_month,String selected_day,String range) throws Exception
	{
		LogData log = LogDataServer.sharedLogServer().get(
			Integer.parseInt(selected_year),
			Integer.parseInt(selected_month),
			Integer.parseInt(selected_day));
		
		/*int log_size = 0;
		for (Enumeration e = log.elements(); e.hasMoreElements(); ) {
			Hashtable bottle = (Hashtable)e.nextElement();
			if ("OB".equals(bottle.get("type"))) {
				log_size++;
			}
		}*/
		int log_size = log.size();
		
		String date = selected_year+"/"+selected_month+"/"+selected_day;
		
		TextTemplate tmpl_main = Httpd.tmpl_index.get("main");
		tmpl_main.set("title",date);
		tmpl_main.set("base","<base href=\"/list/"+date+"/\">");
		tmpl_main.set("date",date);
		
		TextTemplate tmpl_replay_form = Httpd.tmpl_index.get("replay-form");
		tmpl_replay_form.set("year",selected_year);
		tmpl_replay_form.set("month",selected_month);
		tmpl_replay_form.set("day",selected_day);
		tmpl_main.set("replay-form",tmpl_replay_form.evaluate());
		
		TextTemplate tmpl_range = Httpd.tmpl_index.get("range");
		TextTemplate tmpl_splitted_range = Httpd.tmpl_index.get("splitted-range");
		StringBuffer range_html = new StringBuffer();
		for (int from = 1; from <= log_size; from += RANGE_STEP) {
			tmpl_splitted_range.set("from",Integer.toString(from));
			
			if (from + RANGE_STEP <= log_size) {
				tmpl_splitted_range.set("to",Integer.toString(from + RANGE_STEP - 1));
			}
			else {
				tmpl_splitted_range.set("to","");
			}
			
			range_html.append(tmpl_splitted_range.evaluate()+"\n");
		}
		tmpl_range.set("splitted-ranges",range_html.toString());
		tmpl_main.set("range",tmpl_range.evaluate());
		
		int range_start = 1;
		int range_end = -1;
		if (range == null) {
			
		}
		else if (range.startsWith("l")) {
			try {
				range_start = log_size - Integer.parseInt(range.substring(1));
			}
			catch (Exception e) {
				range_start = -1;
			}
		}
		else if (range.indexOf('-') != -1) {
			int minus_pos = range.indexOf('-');
			
			try {
				range_start = Integer.parseInt(range.substring(0,minus_pos));
			}
			catch (Exception e) {}
			
			try {
				range_end = Integer.parseInt(range.substring(minus_pos+1));
			}
			catch (Exception e) {}
		}
		else {
			try {
				range_start = range_end = Integer.parseInt(range);
			}
			catch (Exception e) {}
		}
		
		StringBuffer bottles = new StringBuffer();
		TextTemplate tmpl_bottles = Httpd.tmpl_index.get("bottles");
		TextTemplate tmpl_bottle = Httpd.tmpl_index.get("bottle");
		int i = 1;
		for (Enumeration e = log.elements(); e.hasMoreElements(); i++) {
			Hashtable bottle = (Hashtable)e.nextElement();
			
			if (i < range_start ||
				(range_end != -1 && i > range_end)) {
				
				continue;
			}
			
			/*if (!"OB".equals(bottle.get("type"))) {
				i--;
				continue;
			}*/
			
			tmpl_bottle.set("id",Integer.toString(i));
			tmpl_bottle.set("timestamp",parse_timestamp((String)bottle.get("time")));
			tmpl_bottle.set("channel",(String)bottle.get("channel"));
			tmpl_bottle.set("ghost",format_ghost((String)bottle.get("ifghost")));
			tmpl_bottle.set("n-agree",format_n_agree((String)bottle.get("agree")));
			tmpl_bottle.set("n-vote",format_n_vote((String)bottle.get("vote")));
			tmpl_bottle.set("mid",(String)bottle.get("mid"));
			tmpl_bottle.set("script",format_script((String)bottle.get("script")));
			bottles.append(tmpl_bottle.evaluate()+"\n");
		}
		tmpl_bottles.set("bottles",bottles.toString());
		tmpl_main.set("main-content",tmpl_bottles.evaluate());
		
		ok_200(tmpl_main.evaluate());
	}
	
	public void cmd_vote(StringTokenizer path_st) throws Exception
	{
		String mid = path_st.nextToken();
		
		ServerConnection connection = Foundation.sharedFoundation().getSlppConnection();
		if (connection == null) {
			ok_200("BotCli is currently offline.","text/plain");
		}
		else {
			connection.vote(mid,"Vote");
		}
	}
	
	public void cmd_agree(StringTokenizer path_st) throws Exception
	{
		String mid = path_st.nextToken();
		
		ServerConnection connection = Foundation.sharedFoundation().getSlppConnection();
		if (connection == null) {
			ok_200("BotCli is currently offline.","text/plain");
		}
		else {
			connection.vote(mid,"Agree");
		}
	}
	
	public void cmd_replay(StringTokenizer path_st,String query) throws Exception
	{
		String head_of_path = path_st.nextToken();
		if (head_of_path.equals("mid")) {
			String mid = path_st.nextToken();
			Calendar timestamp = MidAnalyzer.analyzeTimeStampFromMid(mid);
			int year = timestamp.get(Calendar.YEAR);
			int month = timestamp.get(Calendar.MONTH) + 1;
			int day = timestamp.get(Calendar.DAY_OF_MONTH);
			LogData log = LogDataServer.sharedLogServer().get(year,month,day);
			
			if (log == null) {
				err_404();
				return;
			}
			
			boolean found = false;
			for (Enumeration e = log.elements(); e.hasMoreElements(); ) {
				Hashtable bottle = (Hashtable)e.nextElement();
				if (mid.equals((String)bottle.get("mid"))) {
					LogBrowserController.replay_bottle(bottle);
					found = true;
					break;
				}
			}
			if (!found) {
				err_404();
				return;
			}
		}
		else {
			String year = head_of_path;
			String month = path_st.nextToken();
			String day = path_st.nextToken();
			
			Hashtable query_table = parse_query(query);
			int from = Integer.parseInt((String)query_table.get("from"));
			int to = Integer.parseInt((String)query_table.get("to"));
			
			LogData log = LogDataServer.sharedLogServer().get(
				Integer.parseInt(year),
				Integer.parseInt(month),
				Integer.parseInt(day));
				
			/*int log_size = 0;
			for (Enumeration e = log.elements(); e.hasMoreElements(); ) {
				Hashtable bottle = (Hashtable)e.nextElement();
				if ("OB".equals(bottle.get("type"))) {
					log_size++;
				}
			}*/
			int log_size = log.size();
			
			int i = 1;
			for (Enumeration e = log.elements(); e.hasMoreElements(); i++) {
				Hashtable bottle = (Hashtable)e.nextElement();
				
				if (i < from ||
					(to != -1 && i > to)) {
					
					continue;
				}
				
				/*if (!"OB".equals(bottle.get("type"))) {
					i--;
					continue;
				}*/
				
				LogBrowserController.replay_bottle(bottle);
			}
		}
	}
	
	public static Hashtable parse_query(String query)
	{
		Hashtable result = new Hashtable();
		StringTokenizer elem_st = new StringTokenizer(query,"&");
		while (elem_st.hasMoreTokens()) {
			String elem = elem_st.nextToken();
			
			int equal_pos = elem.indexOf('=');
			String key = elem.substring(0,equal_pos);
			String value = elem.substring(equal_pos+1);
			
			result.put(key,value);
		}
		return result;
	}
	
	public static String parse_timestamp(String time)
	{
		if (time == null) {
			return "";
		}
		else {
			return time.substring(0,2)+':'+time.substring(2,4)+':'+time.substring(4);
		}
	}
	
	public static String format_ghost(String ghost)
	{
		if (ghost == null) {
			return "";
		}
		else {
			return "[Ghost: "+ghost+"]";
		}
	}
	
	public static String format_n_agree(String n_agree)
	{
		if (n_agree == null || n_agree.equals("0")) {
			return "";
		}
		else {
			return "[Agree: "+n_agree+"]";
		}
	}
	
	public static String format_n_vote(String n_vote)
	{
		if (n_vote == null || n_vote.equals("0")) {
			return "";
		}
		else {
			return "[Vote: "+n_vote+"]";
		}
	}
	
	public static String format_script(String script)
	{
		return script;
	}
	
	public void cmd_css(StringTokenizer path_st) throws Exception
	{
		int n_tokens = path_st.countTokens();
		if (n_tokens == 0) {
			err_404();
		}
		else {
			String css_name = path_st.nextToken();
			if (css_name.equals("default.css")) {
				ok_200(Httpd.css_default,"text/css; charset=UTF-8");
			}
			else {
				err_404();
			}
		}
	}
	
	public void ok_200(String content) throws Exception
	{
		ok_200(content,null);
	}
	public void ok_200(String content,String content_type) throws Exception
	{
		if (content_type == null) {
			content_type = "text/html; charset=UTF-8";
		}
		
		out.write("HTTP/1.0 200 OK\r\n");
		out.write("Content-Type: "+content_type+"\r\n");
		out.write("\r\n");
		out.write(content);
		out.flush();
	}
	
	public void redirect_found(String location) throws Exception
	{
		out.write("HTTP/1.0 302 Found\r\n");
		out.write("Location: "+location+"\r\n");
		out.write("\r\n");
		out.flush();
	}
	
	public void err_404() throws Exception
	{
		out.write("HTTP/1.0 404 Not Found\r\n");
		out.write("Content-Type: text/plain\r\n");
		out.write("Content-Length: 13\r\n");
		out.write("\r\n");
		out.write("404 Not Found");
		out.flush();
	}
}
